/*
    AVFS: A Virtual File System Library
    Copyright (C) 1998-1999  Miklos Szeredi (mszeredi@inf.bme.hu)

    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 
*/

#ifndef AV_VDEV_H
#define AV_VDEV_H

#include <errno.h>

/* AVFS is modular, so from now on the order of the fields in the structures
   is _FROZEN_. */

#define DIR_SEP_CHAR '/'
#define AVFS_FLAG_NOSIZE (1 << 0)

#define AVFS_MAX_NAMELEN 256
#define AVFS_MAX_PATHLEN 1024

#define VDEV_FLAG_ISDIR (1 << 0)

/* FIXME: Not nice */
#define AVFS_DIRENT_SIZE (AVFS_MAX_NAMELEN + 20)

typedef unsigned char       avbyte;      /* 1 byte unsigned    */
typedef signed char         avsbyte;     /* 1 byte signed      */
typedef unsigned short      avdbyte;     /* 2 bytes unsigned   */
typedef short               avdchar;     /* 2 bytes signed     */
typedef unsigned int        avqbyte;     /* 4 bytes unsigned   */
typedef int                 avqchar;     /* 4 bytes signed     */
typedef unsigned long       avulong;     /* 4-8 bytes unsigned */
typedef long                avlong;      /* 4-8 bytes signed   */

#if 0
typedef long long           avlonglong;  /* 8 bytes signed     */
typedef unsigned long long  avulonglong; /* 8 bytes unsigned   */
#endif

/* Note: these types only affect the _virtual_ files. 
   Let me know if they are not large enough for something */

typedef avqbyte avdev_t;
typedef avdbyte avmode_t;
typedef avdbyte avnlink_t;
typedef avdbyte avuid_t;
typedef avdbyte avgid_t;
typedef avulong avino_t;
typedef avlong  avoff_t;
typedef avlong  avtime_t;
typedef avulong avsize_t;
typedef avlong  avssize_t;
typedef avqbyte avblksize_t;
typedef avqbyte avblkcnt_t;

typedef struct {
  int errn;
} ave;


struct pred_dev {
  int     next;
  int     prev;
  int     minor;
  avdev_t dev;
  avino_t ino;
};

struct ext_info {
  const char *from;
  const char *to;
};

struct avstat {
  avdev_t          dev;
  avino_t          ino;
  avmode_t         mode;
  avnlink_t        nlink;
  avuid_t          uid;
  avgid_t          gid;
  avdev_t          rdev;
  avoff_t          size;
  avblksize_t      blksize;
  avblkcnt_t       blocks;
  avtime_t         atime;
  avtime_t         mtime;
  avtime_t         ctime;
  char             _reserved[32];
};

struct avdirent {
  avino_t       ino;
  avoff_t       off;
  int           reclen;
  int           type;
  char          _reserved[32];
  char          name[AVFS_MAX_NAMELEN+1];
};

struct avutimbuf {
  avtime_t actime;
  avtime_t modtime;
};

struct avtimeval {
  avlong sec;
  avlong usec;
};

#define BASE(path)  __av_get_base(path)
#define EXTRA(path) __av_get_extra(path)

struct _vpath;
typedef struct _vpath vpath;

/* Defined in resolve.h */
struct workpath;

/* This structure contains all the data for a handler */

struct vdev_info {
  void *devdata;

  char *name;
  struct ext_info *exts;

  struct pred_dev *minors;
  int flags;

  void      (*destroy)  (struct vdev_info *di);

  /* If NULL, VFS will resolve the path */
  int       (*resolve)  (ave *v, void *devdata, vpath *path, 
			 char *ipath, int *atp,  struct workpath *wp);

  void *    (*open)     (ave *v, void *devdata, vpath *path, int flags, 
			 int mode);
  int       (*close)    (ave *v, void *devinfo);
  avssize_t (*read)     (ave *v, void *devinfo, char *buf, avsize_t nbyte);
  avssize_t (*write)    (ave *v, void *devinfo, const char *buf, 
			 avsize_t nbyte);
  avoff_t   (*lseek)    (ave *v, void *devinfo, avoff_t offset, int whence);
  int       (*fstat)    (ave *v, void *devinfo, struct avstat *buf, int flags);
  int       (*readdir)  (ave *v, void *devinfo, struct avdirent *ent);

  /* stat can be NULL, because it can be substituted with fstat */
  int       (*stat)    (ave *v, void *devdata, vpath *path,
			struct avstat *buf, int flags, int deref);

  int       (*access)   (ave *v, void *devdata, vpath *path, int amode);

  /* readlink can be NULL if there are no symlinks */
  int       (*readlink) (ave *v, void *devdata, vpath *path,
                        char *buf, avsize_t bufsize);

  int       (*unlink)   (ave *v, void *devdata, vpath *path);
  int       (*rmdir)    (ave *v, void *devdata, vpath *path);
  int       (*mkdir)    (ave *v, void *devdata, vpath *path, avmode_t mode);
  int       (*mknod)    (ave *v, void *devdata, vpath *path, avmode_t mode, 
			 avdev_t dev);
  int       (*rename)   (ave *v, void *devdata, const char *path,
			 vpath *newpath);
  int       (*link)     (ave *v, void *devdata, const char *path,
			 vpath *newpath);
  int       (*symlink)  (ave *v, void *devdata, const char *path,
			 vpath *newpath);

  int       (*chmod)    (ave *v, void *devdata, vpath *path, avmode_t mode);
  int       (*chown)    (ave *v, void *devdata, vpath *path, avuid_t uid,
			 avgid_t gid, int deref);
  int       (*utime)    (ave *v, void *devdata, vpath *path, 
			 struct avutimbuf *buf);
  int       (*fchmod)   (ave *v, void *devinfo, avmode_t mode);
  int       (*fchown)   (ave *v, void *devinfo, avuid_t uid, avgid_t gid);

  int       (*fgetfile)  (ave *v, void *devinfo, avoff_t offset,
			  avsize_t *lenp, avoff_t *retoffp);
};


struct avtm {
  int sec;    /* [0-59] (note: 61 _can_ happen) */
  int min;    /* [0-59] */
  int hour;   /* [0-23] */
  int day;    /* [1-31] */
  int mon;    /* [0-11] */
  int year;   /* Year - 1900 */
};

struct proginfo {
  const char *const *prog;

  int ifd;
  int ofd;
  int efd;
  
  int pid;

  const char *wd;
  char  _reserved[32];
};

typedef struct {
  char *name;
  int is_tmp;
} real_file;


#define AV_NEW(v, ptr) ptr = __av_malloc(v, sizeof(*ptr))
#define DUMMYV (&__av_dummyv)
#define AV_MAX(x, y) ((x) > (y) ? (x) : (y))
#define AV_MIN(x, y) ((x) < (y) ? (x) : (y))
#define AV_DIV(x, y) ((x) ? (((x) - 1) / (y) + 1) : 0)

#define AVNULL 0

#define AVSEEK_SET   0
#define AVSEEK_CUR   1
#define AVSEEK_END   2

#define AVR_OK       4
#define AVW_OK       2
#define AVX_OK       1
#define AVF_OK       0

#define AVO_ACCMODE  0x03

#define AVO_RDONLY   0
#define AVO_WRONLY   1
#define AVO_RDWR     2

#define AVO_CREAT    0100
#define AVO_EXCL     0200
#define AVO_TRUNC    01000
#define AVO_APPEND   02000

#define AV_IFMT   0170000
#define AV_IFDIR  0040000
#define AV_IFCHR  0020000
#define AV_IFBLK  0060000
#define AV_IFREG  0100000
#define AV_IFIFO  0010000
#define AV_IFLNK  0120000
#define AV_IFSOCK 0140000

#define __AV_ISTYPE(mode, mask)  (((mode) & AV_IFMT) == (mask))
 
#define AV_ISDIR(mode)  __AV_ISTYPE((mode), AV_IFDIR)
#define AV_ISCHR(mode)  __AV_ISTYPE((mode), AV_IFCHR)
#define AV_ISBLK(mode)  __AV_ISTYPE((mode), AV_IFBLK)
#define AV_ISREG(mode)  __AV_ISTYPE((mode), AV_IFREG)
#define AV_ISFIFO(mode) __AV_ISTYPE((mode), AV_IFIFO)
#define AV_ISLNK(mode)  __AV_ISTYPE((mode), AV_IFLNK)
#define AV_ISSOCK(mode) __AV_ISTYPE((mode), AV_IFSOCK)

#define AVLOG_SYSCALL   40
#define AVLOG_DEBUG     30
#define AVLOG_WARNING   20
#define AVLOG_ERROR     10
#define AVLOG_NONE      0

#ifdef AVFS_DEBUG
#define AVDEB(x) __av_log x
#else
#define AVDEB(x) 
#endif

extern ave       __av_dummyv;

extern int        __av_add_vdev(ave *v, struct vdev_info *vdev);
extern struct vdev_info * __av_new_vdev(ave *v);

extern vpath      *__av_get_base(vpath *vp);
extern const char *__av_get_extra(vpath *vp);
extern struct vdev_info *__av_get_vdev(vpath *vp);

extern int         __av_isroot(ave *v, vpath *vp);

extern int        __av_getminor(int major, avdev_t odev, avino_t oino);
extern avdev_t    __av_vmakedev(int major, int minor);
extern avdev_t    __av_mkdev(int major, int minor);
extern void       __av_default_stat(struct avstat *stbuf);

extern void       __av_log(int level, const char *format, ...);
extern void      *__av_malloc(ave *v, avsize_t nbyte);
extern void      *__av_realloc(ave *v, void *ptr, avsize_t nbyte);
extern void       __av_free(void *ptr);
extern char      *__av_strdup(ave *v, const char *s);
extern int        __av_strncasecmp(const char *s1, const char *s2, avsize_t n);
extern void       __av_copy_filename(ave *v, char *dst, const char *src);
extern avoff_t    __av_generic_lseek(ave *v, avoff_t *ptrp, avoff_t size,
				     avoff_t offset, int whence);
extern void       __av_init_proginfo(struct proginfo *pi);
extern int        __av_start_prog(ave *v, struct proginfo *pi);
extern int        __av_wait_prog(ave *v, struct proginfo *pi, int tokill,
				 int check);
extern int        __av_pipe(ave *v, int pip[2]);
extern int        __av_set_nonblock(ave *v, int fd);
extern int        __av_localselect(ave *v, int n, void *readfds,
				   void *writefds, void *exceptfds, 
				   struct avtimeval *timeout);
extern void       __av_closeall(void);
extern int        __av_random(void);
extern void       __av_registerfd(int fd);
extern avtime_t   __av_gettime(void);
extern unsigned int __av_sleep(unsigned int seconds);
extern avtime_t   __av_mktime(struct avtm *tp);
extern char      *__av_get_tmpfile(ave *v);
extern void       __av_del_tmpfile(char *tmpfile);
extern int        __av_get_tmpfd(ave *v);

extern int        __av_open(ave *v, vpath *path, int flags, int mode);
extern int        __av_close(ave *v, int fh);
extern avssize_t  __av_read(ave *v, int fh, char *buf, avsize_t nbyte);
extern avssize_t  __av_write(ave *v, int fh, const char *buf, avsize_t nbyte);
extern avoff_t    __av_lseek(ave *v, int fh, avoff_t offset, int whence);
extern int        __av_fstat(ave *v, int fh, struct avstat *buf, int flags);
extern int        __av_stat(ave *v, vpath *path, struct avstat *buf,
			    int flags);
extern int        __av_access(ave *v, vpath *path, int amode);

extern int        __av_chmod(ave *v, vpath *path, avmode_t mode);
extern int        __av_fchmod(ave *v, int fh, avmode_t mode);
extern int        __av_chown(ave *v, vpath *path, avuid_t uid, avgid_t gid);
extern int        __av_fchown(ave *v, int fh, avuid_t uid, avgid_t gid);
extern int        __av_utime(ave *v, vpath *path, struct avutimbuf *buf);

extern real_file *__av_get_realfile(ave *v, vpath *path);
extern void       __av_free_realfile(real_file *rf);

extern int        __av_fgetfile(ave *v, int fh, avoff_t offset, avsize_t *lenp,
				avoff_t *retoffp);

extern int        __av_get_logfile(ave *v);
extern int        __av_localopen(ave *v, const char *path, int flags,
				 int mode);
extern int        __av_localclose(ave *v, int fh);
extern avssize_t  __av_localread(ave *v, int fh, char *buf, avsize_t nbyte);
extern avssize_t  __av_localwrite(ave *v, int fh, const char *buf,
				 avsize_t nbyte);
extern avoff_t    __av_locallseek(ave *v, int fh, avoff_t offset, int whence);
extern int        __av_localunlink(ave *v, const char *path);
extern int        __av_localmkdir(ave *v, const char *path, avmode_t mode);

#endif /* AV_VDEV_H */

