/*  
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)

    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 

    ARCHIVES module
    Created by David Hanak 1998
*/

#include "archives.h"
#include "resolve.h"

#include <ctype.h>
#include <string.h>

#define BYTES_PER_ENTRY AVFS_DIRENT_SIZE
#define MAX_SYMLINKS 10

arch_inode *__av_make_node(ave *v, struct avstat *initstat)
{
  arch_inode *ino;  
  
  AV_NEW(v, ino);
  if(ino == AVNULL) return AVNULL;

  ino->subdir = AVNULL;
  ino->syml = AVNULL;
  ino->udata = AVNULL;
  ino->st = *initstat;
  ino->st.nlink = 0;
  ino->origst =ino->st;
  
  ino->inoflags = 0;

  ino->offset = -1;
  ino->realsize = -1;
  ino->typeflag = -1;

  return ino;
}

arch_entry *__av_new_entry(ave *v)
{
  arch_entry *ent;

  AV_NEW(v, ent);
  if(ent == AVNULL) return AVNULL;
  
  ent->name = AVNULL;
  ent->entflags = 0;
  ent->ino = AVNULL;
  ent->next = AVNULL;
  ent->prevp = AVNULL;

  return ent;
}

char *__av_split_path(ave *v, char *path)
{
  int i;
  char *name;
  
  for(i = strlen(path)-1; i >= 0 && path[i] == DIR_SEP_CHAR; i--);
  path[i+1] = '\0';
  
  for(; i >= 0 && path[i] != DIR_SEP_CHAR; i--);
  name = __av_strdup(v, path+(i+1));
  if(name == AVNULL) return AVNULL;
  
  path[i+1] = '\0';
  
  return name;
}

static int name_equal(arch_entry *ent, const char *pathname, int seglen)
{
  if((int) strlen(ent->name) == seglen) {
    if(!(ent->entflags & ARCHE_NOCASE)) {
      if(strncmp(ent->name, pathname, seglen) == 0) return 1;
    }
    else {
      if(__av_strncasecmp(ent->name, pathname, seglen) == 0) return 1;
    }
  }
  
  return 0;
}

static arch_entry *path_to_entry(ave *v, arch_inode *root, workpath *path,
				 arch_inode **parentdirp, const char **lastseg,
				 int numsymlinks)

{
  unsigned int segbeg;
  unsigned int seglen;
  int found;
  arch_entry *ent = root->subdir; /* Is this OK? This is '.' */
  arch_entry *newent;
  arch_inode *parentdir = root;
  char *syml;

  while(1) {
    for(; AR_CURR(path) == DIR_SEP_CHAR; AR_STEP(path));

    if(!AR_CURR(path)) break;     /* The parsing is finished, get out */

    if(!AV_ISDIR(root->st.mode)) {
      v->errn = ENOTDIR;
      path->i = -1;
      return AVNULL;
    }
    
    segbeg = path->i;
    
    for(;AR_CURR(path) && AR_CURR(path) != DIR_SEP_CHAR; AR_STEP(path));
    seglen = path->i - segbeg;
    
    found = 0;
    for(newent = root->subdir; newent != AVNULL; newent = newent->next) 
      if(name_equal(newent, path->s + segbeg, seglen)) {
	found = 1;
	break;
      }

    if(!found) {
      for(; AR_CURR(path) == DIR_SEP_CHAR; AR_STEP(path));
      if(!AR_CURR(path)) {	  
	if(lastseg != AVNULL) {
	  *lastseg = (path->s + segbeg);
	  break; 	  /* The parsing is finished, get out */
	}
	else v->errn = ENOENT;
      }
      else v->errn = ENOTDIR;

      path->i = -1;
      return AVNULL;
    }
    
    ent = newent;
    if(AV_ISLNK(ent->ino->st.mode) && (syml = ent->ino->syml) != AVNULL &&
       (path->resolvelast || AR_CURR(path) || !AR_END(path))) {

      if(syml[0] != DIR_SEP_CHAR) {
	workpath wp;

	if(numsymlinks > MAX_SYMLINKS) {
	  v->errn = ELOOP;
	  return AVNULL;
	}
	wp.i = 0;
	wp.s = syml;
	wp.l = 0;
	wp.resolvelast = 1;
	
	ent = path_to_entry(v, root, &wp, AVNULL, lastseg, numsymlinks + 1);
	if(ent == AVNULL) {
	  path->i = -1;
	  return AVNULL;
	}
      }
      else {
	v->errn = ENOENT;
	path->i = -1;
	return AVNULL;
      }
    }

    /* If we stepped out from the root of the archive with '..' */
    if(root == ent->ino && root->subdir->next == ent) {
      v->errn = ENOENT;
      return AVNULL;
    }

    parentdir = root;
    root = ent->ino;
  }

  if(parentdirp != AVNULL) *parentdirp = parentdir;
  return ent;
}

arch_entry *__av_find_entry(ave *v, arch_inode *root, const char *path)
{
  workpath wp;

  wp.i = 0;
  wp.s = path;
  wp.l = 0; /* This is OK, but be careful with AR_END() */
  wp.resolvelast = 0;
  
  return path_to_entry(v, root, &wp, AVNULL, AVNULL, 0);
}

static arch_entry *find_entry(ave *v, archive *arch, const char *path, 
			      const char **lastseg)
{
  arch_entry *ent;
  workpath wp;
  
  /* Discard params */
  while(*path && *path != DIR_SEP_CHAR) path++;
  
  wp.i = 0;
  wp.s = path;
  wp.l = 0; /* This is OK, but be careful with AR_END() */
  wp.resolvelast = 0;
  
  ent = path_to_entry(v, arch->root, &wp, AVNULL, lastseg, 0);
  return ent;
}
			     

static arch_inode *find_inode(ave *v, archive *arch, const char *path)
{
  arch_entry *ent;
  
  ent = find_entry(v, arch, path, AVNULL);
  if(ent != AVNULL) return ent->ino;
  else return AVNULL;
}


/* Watch out! This function doesn't check that you are not unlinking a
   directory containing entries (which would leave unfreed memory)!  */

void __av_unlink_entry(arch_entry *ent)
{
  *ent->prevp = ent->next;
  if (ent->next) ent->next->prevp = ent->prevp;

  __av_free(ent->name);
  if(ent->ino->st.nlink == 1) {
    __av_free(ent->ino->syml);
    __av_free(ent->ino->udata);
    __av_free(ent->ino);
  }
  else ent->ino->st.nlink--;

  __av_free(ent);
}

/* Watch out! This function doesn't check that you are adding an entry to a
   directory, not a regular file (which would lead to unfreed memory
   later)! */

void __av_insert_entry(arch_inode *dir, arch_entry *ent)
{
  arch_entry **ep;

  for(ep = &dir->subdir; *ep != AVNULL; ep = &((*ep)->next));
  ent->prevp = ep;
  ent->next = AVNULL;
  *ep = ent;

  ent->ino->st.nlink++;
}

static void free_tree(arch_inode *dir, arch_inode *parentdir)
{
  arch_entry *ent;
  
  while(dir->subdir) {
    ent = dir->subdir;
    if(ent->ino != dir && ent->ino != parentdir) free_tree(ent->ino, dir);
    __av_unlink_entry(ent);
  }
}

static void delete_tree(arch_inode *root)
{
  if(root != AVNULL) {
    root->st.nlink++;   
    free_tree(root, AVNULL);
    __av_free(root);
  }
}

arch_inode *__av_make_dir(ave *v, arch_inode *parentdir, 
			  struct avstat *initstat)
{
  arch_inode *dir;
  arch_entry *dot, *dotdot;
  char *dotname, *dotdotname;

  dir = __av_make_node(v, initstat);
  dot = __av_new_entry(v);
  dotdot = __av_new_entry(v);
  dotname = __av_strdup(v, ".");
  dotdotname = __av_strdup(v, "..");
  
  if(dir == AVNULL || dot == AVNULL || dotdot == AVNULL ||
     dotname == AVNULL || dotdotname == AVNULL) {
    __av_free(dir);
    __av_free(dot);
    __av_free(dotdot);
    __av_free(dotname);
    __av_free(dotdotname);
    return AVNULL;
  }
  
  dot->ino = dir;
  dot->name = dotname;
  if(parentdir != AVNULL) dotdot->ino = parentdir;
  else dotdot->ino = dir;
  dotdot->name = dotdotname;

  __av_insert_entry(dir, dot);
  __av_insert_entry(dir, dotdot);

  return dir;
}

static void default_stat(archive *arch, struct avstat *initstat)
{ 
  initstat->dev     = arch->dev;
  initstat->ino     = arch->inoctr++;
  initstat->mode    = 0644 | AV_IFREG;
  initstat->uid     = arch->uid;
  initstat->gid     = arch->gid;
  initstat->size    = 0;
  initstat->blksize = arch->dd->blksize;
  initstat->blocks  = 0;
  initstat->ctime   = arch->mtime;
  initstat->atime   = initstat->ctime;
  initstat->mtime   = initstat->ctime;
}

static arch_inode *make_default_dir(ave *v, archive *arch,
				    arch_inode *parentdir)
{
  arch_inode *dir;
  struct avstat dirstat;
  avmode_t dirmode;

  dirmode = (arch->mode & 0777) | AV_IFDIR;
  if (dirmode & 0400) dirmode = dirmode | 0100;
  if (dirmode & 0040) dirmode = dirmode | 0010;
  if (dirmode & 0004) dirmode = dirmode | 0001;

  default_stat(arch, &dirstat);
  dirstat.mode = dirmode;

  dir = __av_make_dir(v, parentdir, &dirstat);
  if(dir == AVNULL) return AVNULL;

  return dir;
}

arch_inode *__av_make_dir_rec(ave *v, archive *arch, arch_inode *root,
			      char *path, int entflags)
{
  char *name = AVNULL;
  arch_entry *ent = AVNULL;
  arch_inode *parentdir;
  
  for(; *path == DIR_SEP_CHAR; path++);
  if(!*path) return root;

  ent = __av_find_entry(DUMMYV, root, path);
  if(ent != AVNULL) {
    if(AV_ISDIR(ent->ino->st.mode)) return ent->ino;
    else __av_unlink_entry(ent);
  }
  
  ent = __av_new_entry(v);
  if(ent != AVNULL &&
     (name = __av_split_path(v, path)) != AVNULL &&
     (parentdir = __av_make_dir_rec(v, arch, root, path, entflags)) != AVNULL 
     &&
     (ent->ino = make_default_dir(v, arch, parentdir)) != AVNULL) {

    ent->name = name;
    ent->entflags |= entflags;
    __av_insert_entry(parentdir, ent);

    return ent->ino;
  }
  
  __av_free(ent);
  __av_free(name);

  return AVNULL;
}

static void archive_free(void *obj)
{
  archive *arch = (archive *) obj;

  delete_tree(arch->root);
  __av_free(arch->udata);
  __av_free(arch);
}

static int archive_valid(cache_obj *cobj, void *param)
{    
  archive *arch = (archive *) cobj->obj;
  arch_devd *dd = arch->dd;
  avtime_t *mtimep = (avtime_t *) param;

  if((dd->valid_time != -1 && 
      __av_gettime() > cobj->lasttime + dd->valid_time) ||
     (!(dd->typeflags & ARCHF_NOFILE) && arch->mtime != *mtimep)) return -1;
  
  return 1;
}


static archive *get_archive(ave *v, void *devdata, vpath *path)
{
  arch_devd *dd = (arch_devd *) devdata;
  struct avstat statbuf;
  cache_obj *cobj;
  archive *arch;
  avtime_t now;
  int minor = -1;

  now = __av_gettime();
  
  if(!(dd->typeflags &  ARCHF_NOFILE)) {
    if(__av_stat(v, BASE(path), &statbuf, AVFS_FLAG_NOSIZE) == -1) 
      return AVNULL;
    minor = __av_getminor(dd->major, statbuf.dev, statbuf.ino);
  }
  else {
    statbuf.mtime = now;
    statbuf.uid = 0;
    statbuf.gid = 0;
    statbuf.mode = 0777;
    
    if(dd->getminor != AVNULL) {
      minor = (*dd->getminor) (v, path);
      if(minor == -1) return AVNULL;
    }
    else minor = 1;
  }

  cobj = __av_cache_find(v, dd->major, minor, archive_valid, 
			 (void *) &statbuf.mtime);
  if(cobj == AVNULL) return AVNULL;
  if(cobj->obj != AVNULL) 
    return (archive *) cobj->obj;

  AV_NEW(v, arch);
  if(arch == AVNULL) {
    __av_cache_op(cobj, COBJ_FREE);
    return AVNULL;
  }

  cobj->obj = (void *) arch;
  cobj->destroy = archive_free;

  if(dd->valid_time == -1) 
    cobj->keep_time = dd->keep_time;
  else
    cobj->keep_time = dd->valid_time;

  arch->cobj = cobj;
  arch->dd = dd;
  arch->udata = AVNULL;
  arch->root = AVNULL;
  arch->mtime = statbuf.mtime;
  arch->minor = minor;
  arch->dev = __av_vmakedev(dd->major, minor);
  arch->inoctr = 1;

  arch->uid = statbuf.uid;
  arch->gid = statbuf.gid;
  arch->mode = statbuf.mode;

  arch->root = make_default_dir(v, arch, AVNULL);
  if(arch->root == AVNULL) {
    __av_cache_op(cobj, COBJ_FREE);
    return AVNULL;
  }

  dd->parsefunc(v, path, arch);

  __av_cache_op(cobj, COBJ_UNLOCK);
  return arch;
}

static void release_archive(archive *arch)
{
  if(arch != AVNULL)
    __av_cache_op(arch->cobj, COBJ_RELEASE);
}

static int arch_stat(ave *v, void *devdata, vpath *path, 
		     struct avstat *buf, int flags, int deref)
{
  archive *arch;
  arch_inode *ino;

  ino = AVNULL;
  arch = get_archive(v, devdata, path);
  if(arch != AVNULL) {
    ino = find_inode(v, arch, EXTRA(path));
    if(ino != AVNULL) *buf = ino->st;
  }
  release_archive(arch);

  return (ino == AVNULL) ? -1 : 0;
}


static int arch_fstat(ave *v, void *devinfo, struct avstat *buf, int flags)
{
  arch_fdi *di = (arch_fdi *) devinfo;

  *buf = di->ino->st;
  return 0;
}

static int arch_chmod(ave *v, void *devdata, vpath *path, avmode_t mode)
{
  arch_devd *dd = (arch_devd *) devdata;
  archive *arch;
  arch_inode *ino;

  if(!(dd->typeflags & ARCHF_WRITABLE)) {
    v->errn = EROFS;
    return -1;
  }

  ino = AVNULL;
  arch = get_archive(v, devdata, path);
  if(arch != AVNULL) {
    ino = find_inode(v, arch, EXTRA(path));
    if(ino != AVNULL) 
      ino->st.mode = (ino->st.mode & AV_IFMT) | (mode & 07777);
  }
  release_archive(arch);

  return (ino == AVNULL) ? -1 : 0;
}


static int arch_fchmod(ave *v, void *devinfo, avmode_t mode)
{
  arch_fdi *di = (arch_fdi *) devinfo;

  if(!(di->arch->dd->typeflags & ARCHF_WRITABLE)) {
    v->errn = EROFS;
    return -1;
  }

  di->ino->st.mode = (di->ino->st.mode & AV_IFMT) | (mode & 07777);
  return 0;
}

static int arch_chown(ave *v, void *devdata, vpath *path, avuid_t uid,
		      avgid_t gid, int deref)
{
  arch_devd *dd = (arch_devd *) devdata;
  archive *arch;
  arch_inode *ino;

  if(!(dd->typeflags & ARCHF_WRITABLE)) {
    v->errn = EROFS;
    return -1;
  }

  ino = AVNULL;
  arch = get_archive(v, devdata, path);
  if(arch != AVNULL) {
    ino = find_inode(v, arch, EXTRA(path));
    if(ino != AVNULL) {
      ino->st.uid = uid;
      ino->st.gid = gid;
    }
  }
  release_archive(arch);

  return (ino == AVNULL) ? -1 : 0;
}


static int arch_fchown(ave *v, void *devinfo, avuid_t uid, avgid_t gid)
{
  arch_fdi *di = (arch_fdi *) devinfo;

  if(!(di->arch->dd->typeflags & ARCHF_WRITABLE)) {
    v->errn = EROFS;
    return -1;
  }

  di->ino->st.uid = uid;
  di->ino->st.gid = gid;
  
  return 0;
}

static int arch_utime(ave *v, void *devdata, vpath *path, 
		      struct avutimbuf *buf)
{
  arch_devd *dd = (arch_devd *) devdata;
  archive *arch;
  arch_inode *ino;

  if(!(dd->typeflags & ARCHF_WRITABLE)) {
    v->errn = EROFS;
    return -1;
  }

  ino = AVNULL;
  arch = get_archive(v, devdata, path);
  if(arch != AVNULL) {
    ino = find_inode(v, arch, EXTRA(path));
    if(ino != AVNULL) {
      if(buf == AVNULL) 
	ino->st.mtime = ino->st.atime = __av_gettime();
      else {
	ino->st.atime = buf->actime;
	ino->st.mtime = buf->modtime;
      }
    }
  }
  release_archive(arch);

  return (ino == AVNULL) ? -1 : 0;
}


static int arch_close(ave *v, void *devinfo)
{
  int res = 0;
  arch_fdi *di = (arch_fdi *)devinfo;
  
  if(di->file.fh != -1) res = __av_close(v, di->file.fh);
  release_archive(di->arch);
  __av_free(di->udata);
  __av_free(di);
  
  return res;
}


static int open_ino(ave *v, arch_fdi *di, const char *lastseg,
		    int flags, int mode)
{
  arch_inode *ino = di->ino;

  if(lastseg != AVNULL) {
    arch_inode *parentdir = di->ino;
    arch_entry *ent;
    struct avstat initstat;
    /* Create new inode */

    default_stat(di->arch, &initstat);

    initstat.mode  = (mode & 07777) | AV_IFREG;
    initstat.ctime = __av_gettime();
    initstat.atime = initstat.ctime;
    initstat.mtime = initstat.ctime;

    ino = di->ino = __av_make_node(v, &initstat);
    if(ino == AVNULL) return -1;
    ino->inoflags |= (INOF_DIRTY | INOF_CREATED);

    ent = __av_new_entry(v);
    if(ent == AVNULL) {
      __av_free(ino);
      return -1;
    }

    ent->ino = ino;
    ent->name = __av_strdup(v, lastseg);
    if(ent->name == AVNULL) {
      __av_free(ent);
      __av_free(ino);
      return -1;
    }
    
    __av_insert_entry(parentdir, ent);
    parentdir->inoflags |= INOF_DIRTY;
  }
  else {
    /* If it existed */
    if(!AV_ISREG(ino->st.mode) && !AV_ISDIR(ino->st.mode)) {
      if(AV_ISLNK(ino->st.mode)) v->errn = ENOENT;
      else v->errn = ENXIO;
      return -1;
    }
    
    if(AV_ISDIR(ino->st.mode) && (flags & AVO_ACCMODE) != AVO_RDONLY) {
      v->errn = EISDIR;
      return -1;
    }

    if((flags & (AVO_CREAT | AVO_EXCL)) == (AVO_CREAT | AVO_EXCL)) {
      v->errn = EEXIST;
      return -1;
    }
  }

  di->ent = AVNULL;
  di->loc = 0;

  if((flags & AVO_TRUNC) && ino->st.size != 0) {
    ino->st.size = 0;
    ino->inoflags |= INOF_DIRTY;
  }

  di->file.ptr = 0;
  di->ptr = 0;
  di->lptr = 0;
  di->udata = AVNULL;
  di->vdev = di->arch->dd->vdev;
  
  di->offset = ino->offset;
  di->size = ino->realsize;

  if(flags & AVO_APPEND) {
    di->append = 1;
    di->ptr = di->size;
  }
  else di->append = 0;

  return 0;
}


static void *arch_open(ave *v, void *devdata, vpath *path, int flags,
		       int mode)
{
  arch_devd *dd = (arch_devd *) devdata;
  arch_fdi *di = AVNULL;
  int res = -1;

  if(!(dd->typeflags & ARCHF_WRITABLE) && 
     (flags & (AVO_ACCMODE | AVO_CREAT | AVO_TRUNC))) {
    v->errn = EROFS;
    return AVNULL;
  }

  AV_NEW(v, di);
  if(di == AVNULL) return AVNULL;

  di->file.fh = -1;
  di->udata = AVNULL;
  
  di->arch = get_archive(v, devdata, path);

  if(di->arch != AVNULL) {
    const char *lastseg = AVNULL;
    const char **lastsegp;
    arch_entry *ent;

    if(flags & AVO_CREAT) lastsegp = &lastseg;
    else lastsegp = AVNULL;

    ent = find_entry(v, di->arch, EXTRA(path), lastsegp);
    if(ent != AVNULL) {
      if(!(dd->typeflags & ARCHF_NOFILE)) {
	int baseflags;
	
	if((flags & AVO_ACCMODE) == AVO_RDONLY) baseflags = AVO_RDONLY;
	else baseflags = AVO_RDWR;
	
	res = di->file.fh = __av_open(v, BASE(path), baseflags, 0);
      }
      else {
	di->file.fh = -1;
	res = 0;
      }
      
      di->ino = ent->ino;
      if(res != -1) res = open_ino(v, di, lastseg, flags, mode);
    }
  }

  if(res == -1) {
    arch_close(DUMMYV, (void *) di);
    return AVNULL;
  }
  
  return (void *) di;
}


static int arch_unlink(ave *v, void *devdata, vpath *path)
{
  arch_devd *dd = (arch_devd *) devdata;
  archive *arch;
  arch_entry *ent;
  int res = -1;
  
  if(!(dd->typeflags & ARCHF_WRITABLE)) {
    v->errn = EROFS;
    return -1;
  }
  
  arch = get_archive(v, devdata, path);
  if(arch != AVNULL) {
    ent = find_entry(v, arch, EXTRA(path), AVNULL);
    if(ent != AVNULL) {
      if(AV_ISDIR(ent->ino->st.mode)) v->errn = EISDIR;
      else {
	__av_unlink_entry(ent);
	res = 0;
      }
    }
  }
  release_archive(arch);
  
  return res;
}

static int arch_rmdir(ave *v, void *devdata, vpath *path)
{
  arch_devd *dd = (arch_devd *) devdata;
  archive *arch;
  arch_entry *ent;
  int res = -1;
  
  if(!(dd->typeflags & ARCHF_WRITABLE)) {
    v->errn = EROFS;
    return -1;
  }
  
  arch = get_archive(v, devdata, path);
  if(arch != AVNULL) {
    ent = find_entry(v, arch, EXTRA(path), AVNULL);
    if(ent != AVNULL) {
      if(!AV_ISDIR(ent->ino->st.mode)) 
	v->errn = ENOTDIR;
      else if(ent->ino == arch->root) 
	v->errn = EBUSY; /* FIXME: not correct */
      else if(ent->ino->subdir->next->next != AVNULL) 
	v->errn = ENOTEMPTY;
      else {
	__av_unlink_entry(ent->ino->subdir->next); /* .. */
	__av_unlink_entry(ent->ino->subdir);       /* . */
	__av_unlink_entry(ent);
	res = 0;
      }
    }
  }
  release_archive(arch);
  
  return res;
}

static int make_new_directory(ave *v, archive *arch, arch_inode *parentdir,
			      const char *name, avmode_t mode)
{
  struct avstat initstat;
  arch_entry *ent;
  
  ent = __av_new_entry(v);
  if(ent == AVNULL) return -1;
  
  ent->name = __av_strdup(v, name);
  if(ent->name == AVNULL) {
    __av_free(ent);
    return -1;
  }

  default_stat(arch, &initstat);
  
  initstat.mode  = (mode & 07777) | AV_IFDIR;
  initstat.ctime = __av_gettime();
  initstat.atime = initstat.ctime;
  initstat.mtime = initstat.ctime;
  
  ent->ino = __av_make_dir(v, parentdir, &initstat);

  if(ent->ino == AVNULL) {
    __av_free(ent->name);
    __av_free(ent);
    return -1;
  }
  
  __av_insert_entry(parentdir, ent);

  ent->ino->inoflags |= (INOF_DIRTY | INOF_CREATED);
  parentdir->inoflags |= INOF_DIRTY;

  return 0;
}

static int arch_mkdir(ave *v, void *devdata, vpath *path, avmode_t mode)
{
  arch_devd *dd = (arch_devd *) devdata;
  archive *arch;
  arch_entry *ent;
  const char *lastseg = AVNULL;
  int res = -1;
  
  if(!(dd->typeflags & ARCHF_WRITABLE)) {
    v->errn = EROFS;
    return -1;
  }
  
  arch = get_archive(v, devdata, path);
  if(arch != AVNULL) {
    ent = find_entry(v, arch, EXTRA(path), &lastseg);
    if(ent != AVNULL) {
      if(lastseg == AVNULL)
	v->errn = EEXIST;
      else 
	res = make_new_directory(v, arch, ent->ino, lastseg, mode);
    }
  }

  release_archive(arch);
  
  return res;
}


static int arch_readdir(ave *v, void *devinfo, struct avdirent *ent)
{
  arch_fdi *di = (arch_fdi *)devinfo;
  int curr;

  if(!AV_ISDIR(di->ino->st.mode)) {
    v->errn = ENOTDIR;
    return -1;
  }

  curr = di->ptr / BYTES_PER_ENTRY;

  if(!curr || di->loc > curr) {
    curr = 0;
    di->ent = di->ino->subdir;
  }
  
  for(; curr < di->loc && di->ent != AVNULL; di->ent = di->ent->next, curr++);
  di->ptr = curr * BYTES_PER_ENTRY;

  if(di->ent == AVNULL) return 0;

  __av_copy_filename(v, ent->name, di->ent->name);
  ent->ino = di->ent->ino->st.ino;
  di->ent = di->ent->next;
  curr++;
  di->loc = curr;
  di->ptr += BYTES_PER_ENTRY;
  ent->off = di->ptr;
  ent->reclen = BYTES_PER_ENTRY;
  
  return BYTES_PER_ENTRY;
}

static int arch_access(ave *v, void *devdata, vpath *path, int amode)
{
  arch_devd *dd = (arch_devd *) devdata;
  archive *arch;
  arch_inode *ino;
  int res;
  
  if(!(dd->typeflags & ARCHF_WRITABLE) && (amode & AVW_OK)) {
    v->errn = EACCES;
    return -1;
  }
  
  if(!(dd->typeflags & ARCHF_NOFILE)) {
    res = __av_access(v, BASE(path), amode & ~(AVX_OK));
    if(res == -1) return -1;
  }

  ino = AVNULL;
  arch = get_archive(v, devdata, path);
  if(arch != AVNULL) ino = find_inode(v, arch, EXTRA(path));
  
  if(ino != AVNULL) {
    if(((amode & AVW_OK) && !(ino->st.mode & 0222)) || 
       ((amode & AVX_OK) && !AV_ISDIR(ino->st.mode))) {
      v->errn = EACCES;
      res =  -1;
    }
    else res = 0;
  }
  else res = -1;
  release_archive(arch);
  
  return res;
}

static avoff_t arch_lseek(ave *v, void *devinfo, avoff_t offset, int whence)
{
  arch_fdi *di = (arch_fdi *)devinfo;

  return  __av_generic_lseek(v, &di->ptr, di->size, offset, whence);
}

static int arch_readlink(ave *v, void *devdata, vpath *path, 
			 char *buf, avsize_t bufsize)
{
  archive *arch;
  int retval;

  retval = -1;
  arch = get_archive(v, devdata, path);
  if(arch != AVNULL) {
    arch_inode *ino;
    
    ino = find_inode(v, arch, EXTRA(path));
    if(ino != AVNULL){
      if(ino->syml != AVNULL) {
	strncpy(buf, ino->syml, bufsize);
	retval = strlen(ino->syml);
      }
      else v->errn = EINVAL;
    }
  }
  release_archive(arch);

  return retval;
}


static avssize_t arch_read(ave *v, void *devinfo, char *buf, avsize_t nbyte)
{
  arch_fdi *di = (arch_fdi *) devinfo;
  avoff_t offset = di->offset;
  avoff_t size   = di->size;
  int res;
  avoff_t nact;

  if(AV_ISDIR(di->ino->st.mode)) {
    v->errn = EISDIR;
    return -1;
  }

  if(di->ptr >= size) return 0;
  
  if(offset + di->ptr != di->file.ptr) {
    di->file.ptr = __av_lseek(v, di->file.fh, offset + di->ptr, AVSEEK_SET);
    if(di->file.ptr == -1) return -1;
  }
  nact = AV_MIN(nbyte, (avsize_t) (size - di->ptr));

  res = __av_read(v, di->file.fh, buf, nact);
  if(res == -1) {
    di->file.ptr = -1;
    return -1;
  }
  else {
    di->file.ptr += res;
    if(res != nact) {
      v->errn = EIO;
      return -1;
    }
  }
  di->ptr += nact;

  return nact;
}

static avssize_t arch_write(ave *v, void *devinfo, const char *buf, 
			    avsize_t nbyte)
{
  v->errn = EFAULT;
  return -1;
}

static arch_entry *find_parentdir(arch_inode **parentdirp)
{
  arch_inode *oldpar = *parentdirp;
  arch_inode *newpar;
  arch_entry *ent;
  
  newpar = oldpar->subdir->next->ino; /* This must be '..', mustn't it? */
  *parentdirp = newpar;
  
  for(ent = newpar->subdir; ent != AVNULL; ent = ent->next) 
    if(ent->ino == oldpar) break;
  
  return ent;
}
static int add_to_ipath(ave *v, char *ipath, int *atp, const char *name)
{
  const char *s;
  int at;

  at = *atp;
  ipath[at++] = DIR_SEP_CHAR;
  if(checkl(v, at)) return -1;

  for(s = name; *s; s++) {
    ipath[at++] = *s;
    if(checkl(v, at)) return -1;
  }
  
  *atp = at;
  return 1;
}

static int entry_to_ipath(ave *v, char *ipath, int *atp, arch_inode *parentdir,
			  arch_entry *ent)
{
  int res;
  arch_entry *par_ent;

  if(ent->ino == parentdir) return 1;

  par_ent = find_parentdir(&parentdir);
  if(par_ent == AVNULL) {
    v->errn = EFAULT;
    return -1;
  }

  res = entry_to_ipath(v, ipath, atp, parentdir, par_ent);
  if(res != 1) return res;

  return add_to_ipath(v, ipath, atp, ent->name);
}

static int arch_resolve(ave *v, void *devdata, vpath *path,
			char *ipath, int *atp, workpath *wp)
{
  archive *arch;
  arch_inode *base_ino;
  arch_inode *parentdir;
  arch_entry *ent;
  int ret = -1;
  int isbeg;
  int at;
  const char *lastseg = AVNULL;

  if(*atp == 0) {
    at = *atp;
    isbeg = 1;
    while(AR_CURR(wp) && AR_CURR(wp) != DIR_SEP_CHAR) {
      ipath[at++] = AR_CURR(wp);
      AR_STEP(wp);
      if(checkl(v, at)) return -1;
    }
    while(AR_CURR(wp) == DIR_SEP_CHAR) AR_STEP(wp);
    *atp = at;
    if(!AR_CURR(wp)) return 1;
  }
  else isbeg = 0;

  ipath[*atp] = '\0';
  arch = get_archive(v, devdata, path);
  if(arch != AVNULL) {
    if(!isbeg) {
      base_ino = find_inode(v, arch, ipath);
      
      /* Sadly we must get rid of everything in ipath except the parameters */
      for(at = 0; ipath[at] && ipath[at] != DIR_SEP_CHAR; at++);
      *atp = at;
    }
    else 
      base_ino = arch->root;

    if(base_ino != AVNULL) {
      ent = path_to_entry(v, base_ino, wp, &parentdir, &lastseg, 0);
      if(ent == AVNULL) {
	if(wp->i == -1) ret = -1;
	else ret = 0;
      }
      else {
	if(ent->ino->subdir != AVNULL) {
	  parentdir = ent->ino;
	  ent = find_parentdir(&parentdir);
	}
	ret = entry_to_ipath(v, ipath, atp, parentdir, ent);
	if(ret != -1 && lastseg != AVNULL) 
	  ret = add_to_ipath(v, ipath, atp, lastseg);
      }
    }
    release_archive(arch);
  }
  return ret;
}


static int arch_fgetfile(ave *v, void *devinfo, avoff_t offset, avsize_t *lenp,
			 avoff_t *retoffp)
{
  arch_fdi *di = (arch_fdi *) devinfo;
  
  if(AV_ISDIR(di->ino->st.mode)) {
    v->errn = EISDIR;
    return -1;
  }

  if(offset > di->size) {
    v->errn = EINVAL;
    return -1;
  }

  *lenp = AV_MIN(*lenp, (avsize_t) (di->size - offset));
  
  return __av_fgetfile(v, di->file.fh, di->offset + offset, lenp, retoffp);
}


static void arch_destroy(struct vdev_info *vdev)
{
  arch_devd *dd = (arch_devd *)vdev->devdata;

  __av_free(vdev->name);

  if(dd != AVNULL) {
    __av_free(dd->vdev);
    __av_free(dd->udata);
    __av_cache_freeall(dd->major);
  }

  __av_free(dd);
  __av_free(vdev);
}

struct vdev_info *__av_init_arch(ave *v, const char *name)
{
  arch_devd *dd;
  struct vdev_info *vdev;

  vdev = __av_new_vdev(v);
  if (vdev == AVNULL) return AVNULL;

  AV_NEW(v, dd);
  if((vdev->devdata = (void *) dd) == AVNULL) goto error;
  dd->udata = AVNULL;
  dd->typeflags = 0;
  dd->keep_time = KEEP_TIME;
  dd->valid_time = -1;
  dd->getminor = AVNULL;
  dd->blksize = VBLOCKSIZE;

  AV_NEW(v, dd->vdev);
  if(dd->vdev == AVNULL) goto error;

  dd->vdev->open     = vdev->open     = arch_open;
  dd->vdev->close    = vdev->close    = arch_close;
  dd->vdev->read     = vdev->read     = arch_read;
  dd->vdev->write    = vdev->write    = arch_write;
  dd->vdev->lseek    = vdev->lseek    = arch_lseek;
  dd->vdev->fstat    = vdev->fstat    = arch_fstat;
  dd->vdev->stat     = vdev->stat     = arch_stat;
  dd->vdev->readdir  = vdev->readdir  = arch_readdir;
  dd->vdev->access   = vdev->access   = arch_access;
  dd->vdev->readlink = vdev->readlink = arch_readlink;
  
  dd->vdev->unlink   = vdev->unlink   = arch_unlink;
  dd->vdev->rmdir    = vdev->rmdir    = arch_rmdir;
  dd->vdev->mkdir    = vdev->mkdir    = arch_mkdir;

  dd->vdev->chmod    = vdev->chmod    = arch_chmod;
  dd->vdev->fchmod   = vdev->fchmod   = arch_fchmod;
  dd->vdev->chown    = vdev->chown    = arch_chown;
  dd->vdev->fchown   = vdev->fchown   = arch_fchown;
  dd->vdev->utime    = vdev->utime    = arch_utime;

  dd->vdev->fgetfile = vdev->fgetfile = arch_fgetfile;
  dd->vdev->resolve  = vdev->resolve  = arch_resolve;

  dd->vdev->destroy  = vdev->destroy  = arch_destroy;

  if((vdev->name = __av_strdup(v, name)) == AVNULL) goto error;
  vdev->flags = VDEV_FLAG_ISDIR;

  return vdev;

 error:
  arch_destroy(vdev);
  return AVNULL;
}
