/*
    AVFS: A Virtual File System Library
    Copyright (C) 1998-1999  Miklos Szeredi (mszeredi@inf.bme.hu)

    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 

    CACHE module
*/

#include "cache.h"

#define CACHE_DEFAULT_KEEP_TIME 300 

#define CF_DELETED (1 << 0)
#define CF_LOCKED  (1 << 1)

static cache_obj *first_cobj = AVNULL;
static int cache_lock = 0;

static void remove_cobj(cache_obj *cobj)
{
  *cobj->prevp = cobj->next;
  if(cobj->next) cobj->next->prevp = cobj->prevp;
  
  if(cobj->destroy != AVNULL && cobj->obj != AVNULL) 
    (*cobj->destroy) (cobj->obj);

  __av_free(cobj);
}

static void delete_cobj(cache_obj *cobj)
{
  cobj->state |= CF_DELETED;
}

static void flush_cache(void)
{
  cache_obj *cobj, *nextcobj;  
  avtime_t now;
  int numremoved;
  
  /* FIXME: proper locking */

  if(cache_lock) return;
  cache_lock = 1;
  
  now = __av_gettime();

  do {
    numremoved = 0;
    for(cobj = first_cobj; cobj != AVNULL; cobj = nextcobj) {
      nextcobj = cobj->next;
      
      if(!cobj->usectr && 
	 ((cobj->state & CF_DELETED) || 
	  cobj->lasttime + cobj->keep_time < now)) {
	AVDEB((AVLOG_DEBUG, "Flushing cache object %x/%x age: %i state: %i", 
	       cobj->major, cobj->minor, now - cobj->lasttime, cobj->state));
	
	remove_cobj(cobj);
	numremoved ++;
      }
    }
  } while(numremoved);

  cache_lock = 0;
}

cache_obj *__av_cache_find(ave *v, int major, int minor, cvalid_func valid, 
			   void *param)
{
  cache_obj *cobj, *nextcobj;
  int found;
  
  for(cobj = first_cobj; cobj != AVNULL; cobj = nextcobj) {
    nextcobj = cobj->next;
    
    if(cobj->major == major && cobj->minor == minor) {
      if(valid != AVNULL) found = (*valid) (cobj, param);
      else found = 1;

      if(found == 1) {
	if(cobj->state & CF_LOCKED) {
	  /* FIXME: locking */
	  __av_log(AVLOG_WARNING, "Cannot return locked cache object %x/%x", 
		   cobj->major, cobj->minor);
	}
	else {
	  AVDEB((AVLOG_DEBUG + 5, "Found cache object %x/%x, ctr %i", 
		   cobj->major, cobj->minor, cobj->usectr));
	  
	  cobj->usectr ++;
	  flush_cache();

	  return cobj;
	}
      }
      else if(found == -1) {
	AVDEB((AVLOG_DEBUG, "Invalid cache object %x/%x, ctr %i",
	       cobj->major, cobj->minor, cobj->usectr));
	
	delete_cobj(cobj);
      }
    }
  }

  flush_cache();

  AV_NEW(v, cobj);
  if(cobj == AVNULL) return AVNULL;

  AVDEB((AVLOG_DEBUG, "Creating cache object %x/%x", 
	 major, minor));
  
  cobj->major = major;
  cobj->minor = minor;
  cobj->usectr = 1;
  cobj->state = CF_LOCKED;
  cobj->crtime = cobj->lasttime = __av_gettime();
  cobj->keep_time = CACHE_DEFAULT_KEEP_TIME;
  cobj->destroy = AVNULL;
  cobj->obj = AVNULL;

  cobj->next = first_cobj;
  cobj->prevp = &first_cobj;
  if(first_cobj != AVNULL) first_cobj->prevp = &cobj->next;
  first_cobj = cobj;
  
  return cobj;
}

void __av_cache_op(cache_obj *cobj, int op)
{
  switch(op) {
  case COBJ_FREE:
    cobj->usectr --;    
  case COBJ_DELETE:
    AVDEB((AVLOG_DEBUG + 5, "Deleting cache object %x/%x, ctr %i", 
	   cobj->major, cobj->minor, cobj->usectr));
    delete_cobj(cobj);
  
    flush_cache();
    break;
    
  case COBJ_RELEASE:
    cobj->usectr --;
    cobj->lasttime = __av_gettime();
    AVDEB((AVLOG_DEBUG + 5, "Releasing cache object %x/%x, ctr %i", 
	   cobj->major, cobj->minor, cobj->usectr));

    flush_cache();
    break;

  case COBJ_LOCK:
    if(cobj->state & CF_LOCKED) {
      /* FIXME: locking */
      __av_log(AVLOG_WARNING, "Can't lock locked cache object %x/%x", 
	       cobj->major, cobj->minor);
    }
    else {
      cobj->state |= CF_LOCKED;
      AVDEB((AVLOG_DEBUG + 6, "Locking cache object %x/%x, ctr %i", 
	     cobj->major, cobj->minor, cobj->usectr));
    }
    break;
    
  case COBJ_UNLOCK:
    cobj->state &= ~CF_LOCKED;
    AVDEB((AVLOG_DEBUG + 6, "Unlocking cache object %x/%x, ctr %i", 
	   cobj->major, cobj->minor, cobj->usectr));
    break;
  }
}

void __av_cache_freeall(int major)
{
  cache_obj *cobj, *nextcobj;  

  for(cobj = first_cobj; cobj != AVNULL; cobj = nextcobj) {
    nextcobj = cobj->next;

    if(cobj->major == major) {
      if(cobj->usectr) 
	__av_log(AVLOG_WARNING, 
		 "cache_freeall object with nonzero count: %x/%x, ctr %i", 
		 cobj->major, cobj->minor, cobj->usectr);

      AVDEB((AVLOG_DEBUG, 
	     "cache_freeall cache object: %x/%x", 
	     cobj->major, cobj->minor));
      
      remove_cobj(cobj);
    }
  }
}
