/*
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)

    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 

    OP_DIR module (vfs layer)
*/

#include "op_util.h"


static void convert_dirent(struct avdirent *vent, struct dirent *ent)
{
  ent->d_ino = vent->ino;
  ent->d_off = vent->off;
  ent->d_reclen = vent->reclen;

#ifdef LINUX_GLIBC
  /* FIXME: Ugly hack for GLIBC!!! */
  strcpy(ent->d_name - 1, vent->name);
#else
  strcpy(ent->d_name, vent->name);
#endif
}


static LOCAL3(getdents, int, int, struct dirent *, size_t)

static int common_readdir(ave *v, finfo *fi, struct avdirent *ent, 
			  size_t nbyte)
{
  int res;

  if(fi->vdev->readdir != NULL) {
    res = (*fi->vdev->readdir)(v, fi->devinfo, ent);
    if(res > 0) {
      if(strchr(ent->name, VDEV_SEP_CHAR) != NULL) {
	char *s;
	int i;
	
	s = ent->name;
	for(i = 0; *s && i <= AVFS_MAX_NAMELEN; s++, i++) 
	  if(*s == VDEV_SEP_CHAR) i++;
	
	*s = '\0';
	for(; i >= 0; s--, i--) {
	  ent->name[i] = *s;
	  if(*s == VDEV_SEP_CHAR) ent->name[--i] = *s;
	}
      }
    }
  }
  else {
    v->errn = ENODEV;
    res = -1;
  }

  return res;
}

static int do_getdents(ave *v, int fh, struct avdirent *ent, size_t nbyte)
{
  finfo *fi;
  int res;

  fi = check_fh(v, fh);

  if(fi == NULL) res =  -1;
  else if(nbyte < AVFS_DIRENT_SIZE) {
    v->errn = EINVAL;
    res =  -1;
  }
  else res = common_readdir(v, fi, ent, nbyte);

#if 0
  AVLOGSYS("v_getdents", "", ent->d_ino, res);
#endif

  return res;
}

static int get_virtdir(ave *v, int fh, size_t nbyte, struct avdirent **entp)
{
  finfo *fi;
  int res;
  
  fi = &__av_priv.file_table[fh];
  fi->virtent->off = __av_lseek(v, fh, 0, SEEK_CUR) + VIRTDIRSIZE;
  fi->virtdir = 2;
  
  if(nbyte < VIRTDIRSIZE) {
    v->errn = EINVAL;
    res = -1;
  }
  else {
    res = VIRTDIRSIZE;
    *entp = fi->virtent;
  }
  return res;
}


static int __av_getdents(ave *v, int fh, struct dirent *ent, size_t nbyte)
{
  int res;

  AVLOGSYS("getdents", "", fh, nbyte);

  if(DO_VIRTDIR(fh) && __av_priv.file_table[fh].virtdir == 2) res = 0;
  else {

    if(IS_LOCAL(fh)) res = __av_localgetdents(v, fh, ent, nbyte);
    else {
      struct avdirent vent;
      res = do_getdents(v, fh, &vent, nbyte);
      if(res > 0) convert_dirent(&vent, ent);
    }


    if(DO_VIRTDIR(fh) && res == 0) {
      struct avdirent *virentp;
      res =  get_virtdir(v, fh, nbyte, &virentp);
      if(res != -1) convert_dirent(virentp, ent);
    }
  }

  AVLOGSYS("  (getdents)", "", res, v->errn );

  return res;
}


int __av_virtgetdents(int fh, void *ent, size_t nbyte)
{
  ave v;
  int res;

  v.errn = errno;
  res = __av_getdents(&v, fh, (struct dirent *) ent, nbyte);
  errno = v.errn;
  return res;
}

#ifdef HAVE_OPEN64

struct sun_dirent64 {
  p_ino64_t         d_ino;
  p_off64_t         d_off;
  unsigned short    d_reclen;
  char              d_name[1];
};

static void convert_dirent64(struct avdirent *vent, void *ent)
{
  struct sun_dirent64 *lent = (struct sun_dirent64 *) ent;

  lent->d_ino = vent->ino;
  lent->d_off = vent->off;
  lent->d_reclen = vent->reclen;
  strcpy(lent->d_name, vent->name);
}


static LOCAL3(getdents64, int, int, void *, size_t)

static int __av_getdents64(ave *v, int fh, void *ent, size_t nbyte)
{
  int res;

  AVLOGSYS("getdents64", "", fh, nbyte);

  if(DO_VIRTDIR(fh) && __av_priv.file_table[fh].virtdir == 2) res = 0;
  else {

    if(IS_LOCAL(fh)) res = __av_localgetdents64(v, fh, ent, nbyte);
    else {
      struct avdirent vent;
      res = do_getdents(v, fh, &vent, nbyte);
      if(res > 0) convert_dirent64(&vent, ent);
    }

    if(DO_VIRTDIR(fh) && res == 0) {
      struct avdirent *virentp;
      res =  get_virtdir(v, fh, nbyte, &virentp);
      if(res != -1) convert_dirent64(virentp, ent);
    }
  }

  AVLOGSYS("  (getdents64)", "", res, v->errn );

  return res;
}



int __av_virtgetdents64(int fh, void *ent, size_t nbyte)
{
  ave v;
  int res;

  v.errn = errno;
  res = __av_getdents64(&v, fh, ent, nbyte);
  errno = v.errn;
  return res;
}

#endif /* HAVE_OPEN64 */

#ifdef EMULATE_DIROPS

static LOCAL1(readdir, struct dirent *, DIR *)
static LOCAL1(telldir, off_t, DIR *)

static void __av_localseekdir(ave *v, DIR *dirp, off_t loc)
{
  errno = v->errn;
  __av_realseekdir(dirp, loc);
  v->errn = errno;
}

static void __av_localrewinddir(ave *v, DIR *dirp)
{
  errno = v->errn;
  __av_realrewinddir(dirp);
  v->errn = errno;
}


static off_t __av_telldir(ave *v, AVFS_DIR *dirp)
{
  finfo *fi;

  if(DO_VIRTDIR(dirp->fd) && __av_priv.file_table[dirp->fd].virtdir == 2)
    return __av_priv.file_table[dirp->fd].virtent->off;

  if(IS_LOCAL(dirp->fd)) return __av_localtelldir(v, (DIR *) dirp);

  fi = check_fh(v, dirp->fd);
  if(fi == NULL) return -1;

  return (*fi->vdev->lseek)(v, fi->devinfo, 0, SEEK_CUR);
}

off_t __av_virttelldir(void *dirp)
{
  ave v;
  off_t res;

  v.errn = errno;
  res = __av_telldir(&v, (AVFS_DIR *) dirp);
  errno = v.errn;
  return res;
}


static struct dirent *do_readdir(ave *v, AVFS_DIR *dirp)
{
  finfo *fi;
  int res;
  struct dirent *ent;
  int errn_save;
  struct avdirent vent;

  errn_save = v->errn;

  if(IS_LOCAL(dirp->fd)) return __av_localreaddir(v, (DIR *) dirp);

  fi = check_fh(v, dirp->fd);
  if(fi == NULL) return NULL;

  ent = (struct dirent *) (void *) dirp->buf;

  res = common_readdir(v, fi, &vent, AVFS_MAX_NAMELEN);
  if(res > 0) convert_dirent(&vent, ent);


  AVLOGSYS("readdir", res > 0 ? ent->d_name : "(null)", ent->d_ino, res);

  if(res == -1) return NULL;
  if(res == 0) {
    /* Is this OK? */
    v->errn = errn_save;
    return NULL;
  }

  return ent;
}

static struct dirent *__av_readdir(ave *v, AVFS_DIR *dirp)
{
  struct dirent *ent;

  if(DO_VIRTDIR(dirp->fd) && __av_priv.file_table[dirp->fd].virtdir == 2) {
    v->errn = ENOENT;
    return NULL;
  }

  ent = do_readdir(v, dirp);

  /* Note: can't check for ENOENT, because libc sometimes gives EBADF */
  if(DO_VIRTDIR(dirp->fd) && ent == NULL) {
    finfo *fi;
    
    fi = &__av_priv.file_table[dirp->fd];
    ent = (struct dirent *) (void *) dirp->buf;

    fi->virtent->off = __av_telldir(v, dirp) + VIRTDIRSIZE;
    fi->virtdir = 2;
    convert_dirent(fi->virtent, ent);
  }

  return ent;
}

void *__av_virtreaddir(void *dirp)
{
  ave v;
  struct dirent *ent;

  v.errn = errno;
  ent = __av_readdir(&v, (AVFS_DIR *) dirp);
  errno = v.errn;
  return (void *) ent;
}


static void __av_seekdir(ave *v, AVFS_DIR *dirp, off_t loc)
{
  finfo *fi;

  if(DO_VIRTDIR(dirp->fd)) {
    fi = &__av_priv.file_table[dirp->fd];

    if(fi->virtent->off != -1 && loc == fi->virtent->off) {
      fi->virtdir = 2;
      return;
    }
    else fi->virtdir = 1;
  }

  if(IS_LOCAL(dirp->fd)) return __av_localseekdir(v, (DIR *) dirp, loc);

  fi = check_fh(v, dirp->fd);
  if(fi == NULL) return;

  (*fi->vdev->lseek)(v, fi->devinfo, loc, SEEK_SET);
}

void __av_virtseekdir(void *dirp, off_t loc)
{
  ave v;

  v.errn = errno;
  __av_seekdir(&v, (AVFS_DIR *) dirp, loc);
  errno = v.errn;
}

static void __av_rewinddir(ave *v, AVFS_DIR *dirp)
{
  finfo *fi;

  if(DO_VIRTDIR(dirp->fd)) __av_priv.file_table[dirp->fd].virtdir = 1;
  if(IS_LOCAL(dirp->fd)) return __av_localrewinddir(v, (DIR *) dirp);

  fi = check_fh(v, dirp->fd);
  if(fi == NULL) return;

  (*fi->vdev->lseek)(v, fi->devinfo, 0, SEEK_SET);
}


void __av_virtrewinddir(void *dirp)
{
  ave v;

  v.errn = errno;
  __av_rewinddir(&v, (AVFS_DIR *) dirp);
  errno = v.errn;
}

#endif /* EMULATE DIROPS */
