/*  
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)
    
    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 
*/

#include "config.h"
#include "real.h"

#include "vdev.h"

#include <unistd.h>
#include <sys/syscall.h>

/* #define REALFUNCTYP static inline */
#define REALFUNCTYP

#ifdef LINUX_GLIBC
#define XSTAT_CONVERSION
#define XMKNOD_CONVERSION
#endif

#ifdef XSTAT_CONVERSION
#include "xstatconv.c"
#endif

#undef PATH_T
#define PATH_T const char *

#undef REAL
#define REAL(name) __av_real ## name
#undef SYS
#define SYS(name) SYS_ ## name

#define REALCALL1(name, sc, rettyp, a1t) \
REALFUNCTYP rettyp REAL(name) (a1t a1)               \
{ return (rettyp) syscall(SYS(sc), a1); }

#define REALCALL2(name, sc, rettyp, a1t, a2t) \
REALFUNCTYP rettyp REAL(name) (a1t a1, a2t a2)            \
{ return (rettyp) syscall(SYS(sc), a1, a2); }

#define REALCALL3(name, sc, rettyp, a1t, a2t, a3t) \
REALFUNCTYP rettyp REAL(name) (a1t a1, a2t a2, a3t a3)         \
{ return (rettyp) syscall(SYS(sc), a1, a2, a3); }

#define REALCALL6(name, sc, rettyp, a1t, a2t, a3t, a4t, a5t, a6t)  \
REALFUNCTYP rettyp REAL(name) (a1t a1, a2t a2, a3t a3, a4t a4, a5t a5, a6t a6)\
{ return (rettyp) syscall(SYS(sc), a1, a2, a3, a4, a5, a6); }



REALCALL3(open, open, int, PATH_T, int, int)
REALCALL1(close, close, int, int)

#ifdef HAVE__LIBC_READ
/* We use _libc_{read,write} on Solaris 2.5.1 , because thay have buit in
   restarting, which cannot properly be emulated. */

extern ssize_t _libc_read (int fd, char *buf, size_t nbyte);
extern ssize_t _libc_write (int fd, const char *buf, size_t nbyte);

REALFUNCTYP ssize_t REAL(read) (int fd, char *buf, size_t nbyte)
{ return _libc_read(fd, buf, nbyte); }

REALFUNCTYP ssize_t REAL(write) (int fd, const char *buf, size_t nbyte)
{ return _libc_write(fd, buf, nbyte); }

#else 
REALCALL3(read, read, ssize_t, int, char *, size_t)
REALCALL3(write, write, ssize_t, int, const char *, size_t)
#endif

REALCALL3(lseek, lseek, off_t, int, off_t, int)
REALCALL3(getdents, getdents, int, int, struct dirent *, size_t)
REALCALL2(access, access, int, PATH_T, int)
REALCALL3(readlink, readlink, int, PATH_T, char *, size_t)
REALCALL1(chdir, chdir, int, PATH_T)
REALCALL1(fchdir, fchdir, int, int)
REALCALL3(execve, execve, int, PATH_T, char *const *, char *const *)



#ifndef XSTAT_CONVERSION

#ifndef SYS_prev_stat
REALCALL2(fstat, fstat, int, int, struct stat *)
REALCALL2(stat, stat, int, PATH_T, struct stat *)
REALCALL2(lstat, lstat, int, PATH_T, struct stat *)
#else
REALCALL2(fstat, prev_fstat, int, int, struct stat *)
REALCALL2(stat, prev_stat, int, PATH_T, struct stat *)
REALCALL2(lstat, prev_lstat, int, PATH_T, struct stat *)
#endif

#else 

REALFUNCTYP int REAL(stat) (const char *name, struct stat *buf)
{
  struct kernel_stat kbuf;
  int res;

  res = syscall(SYS(stat), name, &kbuf);
  if (res == 0) res = xstat_conv (&kbuf, buf);

  return res;
}

REALFUNCTYP int REAL(lstat) (const char *name, struct stat *buf)
{
  struct kernel_stat kbuf;
  int res;

  res = syscall(SYS(lstat), name, &kbuf);
  if (res == 0) res = xstat_conv (&kbuf, buf);

  return res;
}

REALFUNCTYP int REAL(fstat) (int fd, struct stat *buf)
{
  struct kernel_stat kbuf;
  int res;

  res = syscall(SYS(fstat), fd, &kbuf);
  if (res == 0) res = xstat_conv (&kbuf, buf);

  return res;
}
#endif /* XSTAT_CONVERSION */

#ifndef XMKNOD_CONVERSION

#ifndef SYS_prev_mknod
REALCALL3(mknod, mknod, int, PATH_T, mode_t, dev_t)
#else
REALCALL3(mknod, prev_mknod, int, PATH_T, mode_t, dev_t)
#endif

#else /* XMKNOD_CONVERSION */
#include <sys/sysmacros.h>

REALFUNCTYP int REAL(mknod) (PATH_T path, mode_t mode, dev_t dev)
{
  unsigned short int k_dev;
  
  k_dev = ((major (dev) & 0xff) << 8) | (minor (dev) & 0xff);
  return syscall(SYS(mknod), path, mode, k_dev);
}

#endif /* XMKNOD_CONVERSION */



REALCALL2(link, link, int, PATH_T, PATH_T)
REALCALL1(unlink, unlink, int, PATH_T)
REALCALL2(chmod, chmod, int, PATH_T, mode_t)
REALCALL2(fchmod, fchmod, int, int, mode_t)
REALCALL3(fchown, fchown, int, int, uid_t, gid_t)
REALCALL1(rmdir, rmdir, int, PATH_T)
REALCALL2(mkdir, mkdir, int, PATH_T, mode_t)
REALCALL2(symlink, symlink, int, PATH_T, PATH_T)
REALCALL2(rename, rename, int, PATH_T, PATH_T)
REALCALL2(utime, utime, int, PATH_T, struct utimbuf *)

#ifdef EMULATE_UTIMES
REALCALL2(utimes, utimes, int, PATH_T, struct timeval *)
#endif


#ifndef LINUX
REALCALL3(chown, chown, int, PATH_T, uid_t, gid_t)

#ifdef HAVE_LCHOWN
REALCALL3(lchown, lchown, int, PATH_T, uid_t, gid_t)
#endif

#else /* LINUX */
#define L_SYS_lchown 16
#define L_SYS_chown  182

REALFUNCTYP int REAL(lchown) (PATH_T path, uid_t uid, gid_t gid)
{ return syscall(L_SYS_lchown, path, uid, gid); }

REALFUNCTYP int REAL(chown) (PATH_T path, uid_t uid, gid_t gid)
{ 
  /* GLIBC-2.0 seems to opt for the simple case: */
  return syscall(L_SYS_lchown, path, uid, gid);
 
#if 0
  int ret;
  ret = syscall(L_SYS_chown, path, uid, gid);
  if(ret == -1 && errno == ENOSYS) 
    ret = syscall(L_SYS_lchown, path, uid, gid);
  return ret;
#endif
}
#endif /* LINUX */

#ifdef HAVE_MMAP
#ifdef SUN_MMAP
#include <sys/mman.h>

/* Add _MAP_NEW to flags on solaris */
REALFUNCTYP void *REAL(mmap) (void *start, size_t len, int prot, int flags,
			      int fd, off_t offset)
{ return (void *) syscall(SYS(mmap), start, len, prot, flags | _MAP_NEW, 
			   fd, offset); }
#  else 
#  ifdef LINUX

/* On linux Syscalls can take only 5 args, so pass the address of the first
   arg */
REALFUNCTYP void *REAL(mmap) (void *start, size_t len, int prot, int flags,
			      int fd, off_t offset)
{ 	
  unsigned long buffer[6];
  
  buffer[0] = (unsigned long)start;
  buffer[1] = (unsigned long)len;
  buffer[2] = (unsigned long)prot;
  buffer[3] = (unsigned long)flags;
  buffer[4] = (unsigned long)fd;
  buffer[5] = (unsigned long)offset;
  return (void *) syscall(SYS(mmap), buffer);
}

#  else
REALCALL6(mmap, mmap, void *, void *, size_t, int, int, int, off_t)
#  endif
#  endif
#endif

#ifdef HAVE_ACL
extern int _acl (PATH_T path, int cmd, int nent, void *aclbuf);

REALFUNCTYP int REAL(acl) (PATH_T path, int cmd, int nent, void *aclbuf)
{ return _acl(path, cmd, nent, aclbuf); }
#endif


#ifdef HAVE_OPEN64
/* Note: All this is for Solaris 2.6. No other OS with x64 calls was tested 
   yet */

REALCALL3(open64, open64, int, PATH_T, int, int)
REALCALL2(fstat64, fstat64, int, int, void *)
REALCALL2(stat64, stat64, int, PATH_T, void *)
REALCALL2(lstat64, lstat64, int, PATH_T, void *)
REALCALL3(getdents64, getdents64, int, int, void *, size_t)

extern offset_t _llseek(int, offset_t, int);
REALFUNCTYP offset_t REAL(lseek64)(int fd, offset_t offset, int whence)
{ return _llseek(fd, (int) offset, whence); }
#if 0
{ return syscall(SYS(llseek), fd, offset, whence); }
#endif


extern void *_mmap64(void *, size_t, int, int, int, offset_t);
REALFUNCTYP void *REAL(mmap64) (void *start, size_t len, int prot, int flags,
				int fd, offset_t offset)
{ return _mmap64(start, len, prot, flags, fd, offset); }
#if 0
{return _syscall(SYS(mmap64), start, len, prot, flags | _MAP_NEW, fd, offset);}
#endif

#endif /* HAVE_OPEN64 */


#ifdef EMULATE_DIROPS

#define DIROP_PREF(x) __libc_ ## x

extern DIR *DIROP_PREF(opendir) (PATH_T);
extern int  DIROP_PREF(closedir) (DIR *);
extern struct dirent *  DIROP_PREF(readdir) (DIR *);
extern off_t DIROP_PREF(telldir) (DIR *);
extern void DIROP_PREF(seekdir) (DIR *, off_t);
extern void DIROP_PREF(rewinddir) (DIR *);

#define REALDIROP1(name, rettyp, a1t) \
REALFUNCTYP rettyp REAL(name) (a1t a1) \
{ return DIROP_PREF(name) (a1); }

REALDIROP1(opendir, DIR *, PATH_T)
REALDIROP1(closedir, int, DIR *)
REALDIROP1(readdir, struct dirent *, DIR *)
REALDIROP1(telldir, off_t, DIR *)

REALFUNCTYP void REAL(seekdir) (DIR *dirp, off_t loc)
{ DIROP_PREF(seekdir) (dirp, loc); }

REALFUNCTYP void REAL(rewinddir) (DIR *dirp)
{ DIROP_PREF(rewinddir) (dirp); }

#endif /* EMULATE_DIROPS */

