/*  
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Pavel Machek  (pavel@ucw.cz)
    
    Distribute either under GPL or under LGPL.
*/

/* How to use this module? Well, basic usage is like /avfs.tgz@a#utar,
 * where /avfs.tgz is essentially any path, @a is postfix saying you
 * want alien filesystem, and #utar is part passed to midnight as
 * is. This works with any filesystem supported by midnight, so you
 * can do cd /@a#ftp:sunsite.ms.mff.cuni.cz/ as well. If you feel this
 * is not too comfortable, you are right. For filesystems listed in
 * this file, it is possible to use shortcuts, so /@ftp:sunsite/ is
 * alias for /@a#ftp:sunsite. Share and enjoy, 			Pavel
 */

#include <dirent.h>
#include <syslog.h>
#include <sys/stat.h>

#include "vdev.h"
#include "mc_vfs.h"
#include "parse.h"

typedef struct {
  vfs *underlying_vfs;
  char *underlying_name;

  /* Logic is: if we have underlying vfs, we can only serve this
   * single vfs, and we will fake name in between path. If not, we'll
   * look up path given to us, and if it contains valid operator,
   * we'll use that vfs else we'll fallback to vfs_local_ops.  */
  int major; 
} alien_devd;

typedef struct {
  struct alien_devd *dd;
  void *handle;
  int isdir;
  char *myname;
  vfs *vfs;
} alien_fdi;

#define SYSLOG( a, b... )
/*#define SYSLOG syslog*/

#define PROLOG \
        alien_fdi *di = (alien_fdi *) devinfo; \
	vfs *me = di->vfs; \
	void *handle = di->handle; \
	int res = 0;

#define OP(x) \
        if (!me->x) { \
	  v->errn = ENODEV; \
	  return -1; \
	}

#define EPILOG \
        if (res == -1) \
	  v->errn = (me->ferrno) (me); \
	return res;

static int alien_close(ave *v, void *devinfo)
{
  PROLOG;
  OP(close);
  if (!di->isdir)
    res = (me->close) (handle);
  else
    res = (me->closedir) (handle);
  __av_free(di->myname);
  di->myname = NULL;
  EPILOG;
}

static int isdir(vfs *me, char *path)
{
  struct stat buf;
  char namebuf[256];

  strcpy(namebuf, path);

  if (!me->stat) 
    return 0;
  if (-1 == (me->stat) (me, namebuf, &buf))
    return 0;
  return (buf.st_mode & S_IFDIR)==S_IFDIR;
}

static vfs *convert_name(alien_devd *dd, char *buf, vpath *path)
{
  char *s = buf;
  const char *t;
  vpath *b;
  vfs *ret = NULL;
  char *extra;

  t = EXTRA(BASE(path));   while(*t) *s++ = *t++;
  t = dd->underlying_name; while(*t) *s++ = *t++;
  t = extra = EXTRA(path); while(*t) *s++ = *t++;
  *s++ = 0;

  syslog( LOG_INFO, "Alien request for '%s'", buf );

  if (dd->underlying_vfs)
    return dd->underlying_vfs;

  if (extra[0] != '#')
    return &vfs_local_ops;
  ret = vfs_type_from_op(extra+1);
  if (!ret)
    return &vfs_local_ops;
  syslog( LOG_INFO, "...resolved on vfs %x '%s'\n", ret, ret->name );
  return ret;
}

static void convert_statbuf(struct stat *buf2, struct avstat *buf)
{
  buf->dev      = buf2->st_dev;
  buf->ino      = buf2->st_ino;
  buf->mode     = buf2->st_mode;
  buf->nlink    = buf2->st_nlink;
  buf->uid      = buf2->st_uid;
  buf->gid      = buf2->st_gid;
  buf->rdev     = buf2->st_rdev;
  buf->size     = buf2->st_size;
  buf->blksize  = buf2->st_blksize;
  buf->blocks   = buf2->st_blocks;
  buf->atime    = buf2->st_atime;
  buf->mtime    = buf2->st_mtime;
  buf->ctime    = buf2->st_ctime;

}

	
static void *alien_open(ave *v, void *devdata, vpath *path, 
		       int flags, int mode)
{
  alien_fdi *di;
  alien_devd *dd = devdata;
  vfs *me;
  char path2[1024];

  SYSLOG( LOG_INFO, "Alien_opendir" );
  me = convert_name(dd, path2, path);

  AV_NEW(v, di);
  if(di == AVNULL) return AVNULL;

  di->dd = (void *) dd;
  di->vfs = me;
  di->myname = __av_strdup(v, path2);
  if (!di->myname) {
    __av_free(di);
    return AVNULL;
  }

  di->isdir = isdir(me, di->myname);
  if (!di->isdir)
    di->handle = (me->open) (me, di->myname, flags, mode);
  else {
    di->handle = (me->opendir) (me, di->myname);
    SYSLOG( LOG_INFO, "Opened alien directory '%s'\n", path2 );
  }
  strcpy(di->myname, path2);

  if (di->handle)
    return (void *) di;

  __av_free(di);
  return AVNULL;
}

static avssize_t alien_read(ave *v, void *devinfo, char *buf, avsize_t nbyte)
{
  PROLOG;
  SYSLOG( LOG_INFO, "Alien_read" );
  OP(read);
  if (di->isdir) {
    syslog( LOG_INFO, "Attempted to open alien directory as a file\n" );
    v->errn = EISDIR;
    return -1;
  }
  res = (me->read) (handle, buf, nbyte);
  EPILOG;
}

static int alien_readdir(ave *v, void *devinfo, struct avdirent *ent)
{
  struct dirent *dir;
  PROLOG;
  SYSLOG( LOG_INFO, "Reading alien direcotry\n" ); 

  OP(readdir);

  dir = (me->readdir) (handle);

  if(!dir)
    return 0;

  __av_copy_filename(v, ent->name, dir->d_name);
  ent->ino = 0;
  ent->off = 0;
  ent->reclen = res = AVFS_DIRENT_SIZE;
  
  return res;
}

static avssize_t alien_write(ave *v, void *devinfo, const char *buf,
			    avsize_t nbyte)
{
  PROLOG;
  SYSLOG(LOG_INFO, "alien_write" );
  OP(write);
  res = (me->write) (handle, (void *) buf, nbyte);
  EPILOG;
}


static avoff_t alien_lseek(ave *v, void *devinfo, avoff_t offset, int whence)
{
  PROLOG;
  SYSLOG(LOG_INFO, "alien_lseek" );
  if(!di->isdir) {
    OP(lseek);
    res = (me->lseek) (handle, offset, whence);
  }
  else {
    OP(seekdir);
    OP(telldir);

    switch(whence) {
    case AVSEEK_SET:
      (me->seekdir) (handle, offset);
      res = (me->telldir) (handle);
      break;
      
    case AVSEEK_CUR:
      res = (me->telldir) (handle);
      if(res != -1 && offset != 0) {
	(me->seekdir) (handle, res + offset);
	res = (me->telldir) (handle);
      }
      break;
      
    case AVSEEK_END:
      v->errn = ENODEV;
      return -1;
      
    default:
      v->errn = EINVAL;
      return -1;
    }
  }
  EPILOG;
}

static int alien_stat(ave *v, void *devdata, vpath *path, struct avstat *buf, 
		      int flags, int deref)
{
  char path2[1024];
  struct stat buf2;
  alien_devd *dd = (alien_devd *) devdata;
  vfs *me;
  int res = 0;
  
  SYSLOG(LOG_INFO, "alien_stat");

  me = convert_name(dd, path2, path);
  OP(stat);
  if(!deref) 
    res = (me->stat) (me, path2, &buf2);
  else
    res = (me->lstat) (me, path2, &buf2);    
  convert_statbuf(&buf2, buf);

  EPILOG;
}

static int alien_fstat(ave *v, void *devinfo, struct avstat *buf, int flags)
{
  struct stat buf2;
  PROLOG;
  SYSLOG(LOG_INFO, "alien_fstat");
  OP(fstat);
  if (di->isdir) {
    char *tmpname;
    syslog(LOG_INFO, "Attempted to fstat directory\n" );
    
    /* Need to duplicate myname, because me->stat modifies it */
    tmpname = __av_strdup(v, di->myname);
    if(tmpname == AVNULL) return -1;

    res = (me->stat) (me, tmpname, &buf2);
    __av_free(tmpname);

  } else res = (me->fstat) (handle, &buf2);
  convert_statbuf(&buf2, buf);
  EPILOG;
}

static int alien_access(ave *v, void *devdata, const char *path, int amode)
{
  v->errn = EACCES;
  return -1;
}

static void alien_destroy(struct vdev_info *filt_vdev)
{
  /*  exit(123);*/
}

static int init_alien(ave *v, char *name, vfs *underlying_vfs, char *underlying_name)
{
  alien_devd *dd;
  struct vdev_info *filt_vdev;
  int res;

  AV_NEW(v, filt_vdev);
  if(filt_vdev == AVNULL) return -1;
  
  filt_vdev->open     = alien_open;
  filt_vdev->close    = alien_close; 
  filt_vdev->read     = alien_read;
  filt_vdev->write    = alien_write;
  filt_vdev->lseek    = alien_lseek;
  filt_vdev->stat     = alien_stat;
  filt_vdev->fstat    = alien_fstat;
  filt_vdev->access   = alien_access;
  filt_vdev->readdir  = alien_readdir;

  filt_vdev->destroy  = alien_destroy;
  filt_vdev->resolve  = AVNULL;

  if((filt_vdev->name = __av_strdup(v, name)) == AVNULL) goto error;

  AV_NEW(v, dd);

  if((filt_vdev->devdata = (void *) dd) == AVNULL) goto error;
  
  res = __av_add_vdev(v, filt_vdev);
  if(res == -1) goto error;

  dd->major = res;
  dd->underlying_vfs = underlying_vfs;
  dd->underlying_name = underlying_name;
  SYSLOG(LOG_INFO, "Alien module: initializing avfs filesystem %s / mc filesystem %s (%x)", name, underlying_name, underlying_vfs);
  return res;

 error:
  alien_destroy(filt_vdev);
  return -1;
}


#define TRY(a) if (a == -1) return -1;

extern int __av_init_module_alien(ave *v);

int __av_init_module_alien(ave *v)
{
  SYSLOG(LOG_INFO, "Alien/%d: calling mc_vfs_init \n", getpid());
  mc_vfs_init();
  TRY( init_alien(v, "a", NULL, "") );
  TRY( init_alien(v, "ftp", &vfs_ftpfs_ops, "#ftp") );
  
  return 0;
}
